import 'dart:convert';
import 'dart:io';

import 'package:dio/dio.dart';
import 'package:flutter/foundation.dart';
import 'package:uuid_type/uuid_type.dart';

import '../../core/core.dart';
import '../../m4e_forms/m4e_forms.dart';
import '../../models/models.dart';


/// Accessing the instantiated `TransactionApi` object
/// is done by passing its type to the GetIt instance
///
/// Example use of `TransactionApi`:
///
/// ```
/// import 'package:m4edart/m4e_sdk_initialize.dart' as sdk;
///
/// sdk.m4e<TransactionApi>().makeTransaction(Transaction transaction);
/// ```
///
///
/// To create an transaction, you can do
/// ```
/// final transaction = Transaction(from: fromWallet, to: toWallet, amount: amount ...)
/// ```
///
///
abstract class TransactionApi {
  /// Initiates a payment transaction by sending transaction (form) to m4e.
  /// Transaction status is `pending` until m4e returns a response which can be `TransactionResult`
  /// or `ServerException` on success and failure cases respectively.
  ///
  /// Throws `ServerException` with code `FAILED_TRANSACTION` if m4e responds with an error which may
  /// be an unauthorized request (user does not have permission to pay into
  /// the receiving wallet), wallet not found, etc.
  ///
  Future<M4eTransactionResult> initiateTransaction(M4eTransactionForm m4eTransactionForm);

  /// Returns m4e authenticated user's `TransactionLogList` from m4e server
  ///
  /// Throws `ServerException` with code `FAILED_TRANSACTION` if m4e responds with an error which may
  /// be an unauthorized request (user does not have permission to pay into
  /// the receiving wallet), wallet not found, etc.
  ///
  Future<M4eTransactionLogList> getUserTransactionLogs(M4eUniqueId userId);

  /// Returns m4e authenticated user's wallet `TransactionLogList` from m4e server
  ///
  /// Throws `ServerException` with code `FAILED_TRANSACTION` if m4e responds with an error which may
  /// be an unauthorized request (user does not have permission to pay into
  /// the receiving wallet), wallet not found, etc.
  ///
  Future<M4eTransactionLogList> getWalletTransactionLogs(M4eUniqueId walletId);
}

class TransactionApiImpl implements TransactionApi {
  final Dio _client;

  TransactionApiImpl({
    @required Dio client,
    @required M4eLoggingInterceptors loggingInterceptor,
    @required M4eHeaderInterceptors headerInterceptor,
  })  : assert(client != null),
        assert(loggingInterceptor != null),
        assert(headerInterceptor != null),
        _client = client {
    _client?.interceptors?.add(loggingInterceptor);
    _client?.interceptors?.add(headerInterceptor);
  }

  Future<M4eTransactionResult> initiateTransaction(M4eTransactionForm transaction) async {
    try {
      assert(transaction != null, 'Transaction should not be null');

      final _response = await _client.put('/transactions/${uuid.v4()}',
          data: json.encode(transaction.toJson()));

      if (_response.statusCode == 201) {
        return M4eTransactionResult.fromJson(_response.data);
      }

      throw M4eExceptionMessages.kFailedTransactionException;
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }
      if (e.type == DioErrorType.RESPONSE) {
        if (e.response.statusCode == 401) {
          // TODO: implement mapping and returning error responses from server
          throw UnimplementedError();
        }
      }

      throw M4eExceptionMessages.kUndefinedException;
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } catch (e) {
      throw M4eExceptionMessages.kUndefinedException;
    }
  }

  @override
  Future<M4eTransactionLogList> getUserTransactionLogs(M4eUniqueId userId) async {
    try {
      final _response = await _client.get('/transaction-log/$userId');

      if (_response.statusCode == 200) {
        return M4eTransactionLogList.fromJson(_response.data);
      }

      throw M4eExceptionMessages.kUndefinedException;
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }
      if (e.type == DioErrorType.RESPONSE) {
        if (e.response.statusCode == 401) {
          // TODO: implement mapping and returning error responses from server
          throw UnimplementedError();
        }
      }

      throw M4eExceptionMessages.kUndefinedException;
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } catch (e) {
      throw M4eExceptionMessages.kUndefinedException;
    }
  }

  @override
  Future<M4eTransactionLogList> getWalletTransactionLogs(M4eUniqueId walletId) async {
    try {
      final _response = await _client.get('/transaction-log/$walletId');

      if (_response.statusCode == 200) {
        return M4eTransactionLogList.fromJson(_response.data);
      }

      throw M4eExceptionMessages.kUndefinedException;
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }
      if (e.type == DioErrorType.RESPONSE) {
        if (e.response.statusCode == 401) {
          // TODO: implement mapping and returning error responses from server
          throw UnimplementedError();
        }
      }

      throw M4eExceptionMessages.kUndefinedException;
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } catch (e) {
      throw M4eExceptionMessages.kUndefinedException;
    }
  }
}
