import 'dart:convert';
import 'dart:io';

import 'package:dio/dio.dart';
import 'package:flutter_secure_storage/flutter_secure_storage.dart' as storage;
import 'package:flutter/foundation.dart';

import '../../core/core.dart';
import '../../models/models.dart';
import '../../m4e_forms/m4e_forms.dart';

/// Accessing the instantiated `UserApi` methods
/// is done by passing its type to the GetIt instance
///
/// Example use of `UserApi`:
/// ```
/// import 'package:m4edart/m4e_sdk_initialize.dart' as sdk;
///
/// sdk.m4e<UserApi>().registerUser(M4eUserForm userForm);
/// ```
///
abstract class UserApi {
  /// Retrives cached user key-pair from cache
  ///
  /// Throws `CacheException` with code `NO_CACHE_DATA`
  /// when no user key-pair is cached
  Future<Object> getUserKeyPairFromCache();

  /// Cache registered user generated key-pair
  ///
  Future<void> cacheUserKeyPair(Object userKeyPair);

  /// Returns `M4eUser` with `user-id`
  ///
  /// Throws `AuthException` with code `NO_AUTH_USER`
  /// when no user is authenticated
  ///
  /// Throws `ServerException` if call to server fails
  ///
  /// Throws `ServerException` with `NO_INTERNET_CONNECTION` error
  /// code device is not connected to the internet
  ///
  Future<M4eUser> getM4eUser(M4eUniqueId userId);

  /// Creates an m4e user account
  ///
  /// Throws `AuthException` with code `NO_AUTH_USER`
  /// when no user is authenticated
  ///
  /// Throws `ServerException` if call to server fails
  ///
  /// Throws `ServerException` with `NO_INTERNET_CONNECTION` error
  /// code device is not connected to the internet
  ///
  Future<M4eUniqueId> registerM4eUser(M4eUniqueId userId, M4eUserForm userForm);
}

class UserApiImpl implements UserApi {
  final Dio _client;
  final storage.FlutterSecureStorage _storage;

  UserApiImpl({
    @required Dio client,
    @required storage.FlutterSecureStorage storage,
    @required M4eLoggingInterceptors loggingInterceptor,
    @required M4eHeaderInterceptors headerInterceptor,
  })  : assert(client != null),
        assert(storage != null),
        assert(loggingInterceptor != null),
        assert(headerInterceptor != null),
        _client = client,
        _storage = storage {
    _client?.interceptors?.add(loggingInterceptor);
    _client?.interceptors?.add(headerInterceptor);
  }

  Future<Object> getUserKeyPairFromCache() {
    // TODO: implement caching and retriving key-pair
  }

  Future<void> cacheUserKeyPair(Object userKeyPair) {
    // TODO: implement caching and retriving key-pair and possibly exporting to external hidden file system
  }

  @override
  Future<M4eUser> getM4eUser(M4eUniqueId userId) async {
    try {
      final _response = await _client.get('/users/${userId.value.toString()}');

      if (_response.statusCode == 200) {
        return M4eUser.fromJson(_response.data);
      }

      throw M4eExceptionMessages.kFailedTransactionException;
    } on DioError catch (e) {
      // print error in console in debug mode
      logger(e);

      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }
      if (e.type == DioErrorType.RESPONSE) {
        if (e.response.statusCode == 401) {
          // TODO: implement mapping and returning error responses from server
          throw UnimplementedError();
        }
      }

      throw M4eExceptionMessages.kUndefinedException;
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } catch (e) {
      throw M4eExceptionMessages.kUndefinedException;
    }
  }

  @override
  Future<M4eUniqueId> registerM4eUser(
      M4eUniqueId userId, M4eUserForm m4eUserForm) async {
    assert(m4eUserForm != null, 'm4eUserForm should not be null');
    try {
      final _response = await _client.put(
        '/users/register/${userId.value.toString()}',
        data: json.encode(m4eUserForm.toJson()),
      );

      if (_response.statusCode == 200) {
        return M4eUniqueId.fromJson(_response.data);
      }

      throw M4eExceptionMessages.kFailedTransactionException;
    } on DioError catch (e) {
      // print error in console in debug mode
      logger(e);

      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }
      if (e.type == DioErrorType.RESPONSE) {
        if (e.response.statusCode == 401) {
          // TODO: implement mapping and returning error responses from server
          throw UnimplementedError();
        }
      }

      throw M4eExceptionMessages.kUndefinedException;
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } catch (e) {
      throw M4eExceptionMessages.kUndefinedException;
    }
  }
}
