import 'dart:convert';
import 'dart:io';

import 'package:dio/dio.dart';
import 'package:flutter/foundation.dart';

import '../../core/core.dart';
import '../../models/models.dart';
import '../../m4e_forms/m4e_forms.dart';

/// Accessing the instantiated `WalletApi` methods
/// is done by passing its type to the GetIt instance
///
/// Example use of `WalletApi`:
/// ```
/// import 'package:m4edart/m4e_sdk_initialize.dart' as sdk;
///
/// sdk.m4e<WalletApi>().fetchUserWallets(userId);
/// ```
///
abstract class WalletApi {
  /// Returns list of wallets for
  /// m4e user with passed-in entity wallet permissions
  Future<M4eWalletList> getUserWallets(
      M4eUniqueId userId, EntityWalletPermissions permissions);

  /// Creates a new wallet definition
  Future<M4eUniqueId> createWallet(
      M4eUniqueId userId, M4eWalletForm m4eWalletForm);
}

class WalletApiImpl implements WalletApi {
  final Dio _client;

  WalletApiImpl({
    @required Dio client,
    @required M4eLoggingInterceptors loggingInterceptor,
    @required M4eHeaderInterceptors headerInterceptor,
  })  : assert(client != null),
        assert(loggingInterceptor != null),
        assert(headerInterceptor != null),
        _client = client {
    // logs and appends authorization headers to requests
    _client.interceptors.add(loggingInterceptor);
    _client.interceptors.add(headerInterceptor);
  }

  /// Returns list of wallets for current authenticated
  /// m4e user with `OWNER` role entity-wallet attributes
  Future<M4eWalletList> getUserWallets(M4eUniqueId userId,
      [EntityWalletPermissions permissions]) async {
    try {
      final _query =
          permissions != null ? '/access?${permissions.toString()}' : '';

      final _response = await _client.get('/wallets/$userId$_query');

      if (_response.statusCode == 200) {
        return M4eWalletList.fromJson(_response.data);
      }

      throw M4eExceptionMessages.kUndefinedException;
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }
      if (e.type == DioErrorType.RESPONSE) {
        if (e.response.statusCode == 401) {
          // TODO: implement mapping and returning error responses from server
          throw UnimplementedError();
        }
      }

      throw M4eExceptionMessages.kUndefinedException;
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } on M4eAuthException catch (_) {
      throw M4eExceptionMessages.kNoAuthUser;
    } catch (e) {
      throw M4eExceptionMessages.kUndefinedException;
    }
  }

  @override
  Future<M4eUniqueId> createWallet(
      M4eUniqueId userId, M4eWalletForm m4eWalletForm) async {
    assert(m4eWalletForm != null, 'm4eWalletForm should not be null');
    try {
      final _response = await _client.put(
        '/wallets/${userId.toString()}/${m4eWalletForm.id.toString()}/${DateTime.now().toTimestampSinceEpoch()}',
        data: json.encode(m4eWalletForm.toJson()),
      );

      if (_response.statusCode == 200) {
        return M4eUniqueId.fromJson(_response.data);
      }

      throw M4eExceptionMessages.kFailedTransactionException;
    } on DioError catch (e) {
      // print error in console in debug mode
      logger(e);

      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }
      if (e.type == DioErrorType.RESPONSE) {
        if (e.response.statusCode == 401) {
          // TODO: implement mapping and returning error responses from server
          throw UnimplementedError();
        }
      }

      throw M4eExceptionMessages.kUndefinedException;
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } catch (e) {
      throw M4eExceptionMessages.kUndefinedException;
    }
  }
}
