import 'package:data_connection_checker/data_connection_checker.dart';

import '../../core/core.dart';
import '../../m4e_sdk_initialize.dart';
import '../../models/models.dart';
import '../../m4e_forms/m4e_forms.dart';
import '../api/apis.dart';

/// Exposes SDK Wallet Api methods
///
/// Accessing the instantiated `m4eWallet` methods
/// is done by:
///
/// Instantiating as a regular object
///
/// ```
/// final _m4eWallet = M4eWallet();
///
/// await _m4eWallet.getUserWallets();
/// ```
///
class M4eWalletService {
  final AuthApi _authApi;
  final WalletApi _walletApi;
  final DataConnectionChecker _connectionChecker;

  M4eWalletService({
    AuthApi authApi,
    WalletApi walletApi,
    DataConnectionChecker connectionChecker,
  })  : _authApi = authApi ?? m4eSDK<AuthApi>(),
        _walletApi = walletApi ?? m4eSDK<WalletApi>(),
        _connectionChecker =
            connectionChecker ?? m4eSDK<DataConnectionChecker>();

  /// Returns current authenticated m4e user `Wallets` from m4e when no [M4eUniqueId] is provided
  ///
  /// Returns m4e user `Wallets` with provided `M4eUniqueId` from m4e
  ///
  /// Example:
  ///
  /// ```
  /// final service = M4eWalletService(); // instantiate M4eWalletService
  ///
  /// final currentUserWallets = service.getWallets(); // returns current authenticated m4e user wallets
  /// ```
  ///
  /// ```
  /// final otherUserWallets = service.getWallets(M4eUniqueId.fromUniqueString(06fa9dcc-03f0-447e-aaad-7b8c4f1f60ee));  // returns m4e-user with [M4eUniqueId] wallets which the current authenticated user has a `SEE` permission on
  /// ```
  ///
  /// Throws [M4eAuthException] if no m4e user is authenticated
  ///
  /// Throws [M4eServerException] if request fails.
  ///
  /// Throws `[ServerException]` with `NO_INTERNET_CONNECTION` error
  /// code when device is not connected to the internet
  ///
  // TODO: re-implement after enquiry
  Future<M4eWalletList> getWallets([M4eUniqueId userId]) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _token = await _authApi.getAuthTokenFromCache();

        final _userId = userId ?? _token.id;
        final _walletPermission = userId == null
            ? EntityWalletPermissions.owner()
            : EntityWalletPermissions.payTo();

        final _wallets =
            await _walletApi.getUserWallets(_userId, _walletPermission);

        return _wallets;
      } catch (e) {
        rethrow;
      }
    }

    throw M4eExceptionMessages.kNoInternetConnectionException;
  }

  /// Creates an m4e user-wallet and returns created wallet [M4eUniqueId]
  ///
  /// Throws [M4eAuthException] if no m4e user is authenticated
  ///
  /// Throws [M4eServerException] if request fails.
  ///
  /// Throws `[ServerException]` with `NO_INTERNET_CONNECTION` error
  /// code device is not connected to the internet
  ///
  Future<M4eUniqueId> createWallet(M4eWalletForm m4eWalletForm) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _token = await _authApi.getAuthTokenFromCache();

        final _walletId =
            await _walletApi.createWallet(_token.id, m4eWalletForm);

        return _walletId;
      } catch (e) {
        rethrow;
      }
    }

    throw M4eExceptionMessages.kNoInternetConnectionException;
  }
}
