import 'package:data_connection_checker/data_connection_checker.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:m4e_sdk/src/core/core.dart';
import 'package:m4e_sdk/src/data/api/apis.dart';
import 'package:m4e_sdk/src/data/services/services.dart';
import 'package:m4e_sdk/src/m4e_forms/m4e_forms.dart';
import 'package:m4e_sdk/src/models/models.dart';
import 'package:matcher/matcher.dart';
import 'package:mockito/mockito.dart';

class MockAuthApi extends Mock implements AuthApi {}

class MockWalletApi extends Mock implements WalletApi {}

class MockDataConnectionChecker extends Mock implements DataConnectionChecker {}

main() {
  M4eWalletService _m4eWalletService;

  MockAuthApi _mockAuthApi;
  MockWalletApi _mockWalletApi;
  MockDataConnectionChecker _mockDataConnectionChecker;

  setUp(() {
    _mockAuthApi = MockAuthApi();
    _mockWalletApi = MockWalletApi();
    _mockDataConnectionChecker = MockDataConnectionChecker();

    _m4eWalletService = M4eWalletService(
      authApi: _mockAuthApi,
      walletApi: _mockWalletApi,
      connectionChecker: _mockDataConnectionChecker,
    );
  });

  // final _userId =
  //     M4eUniqueId.fromUniqueString('46e87882-4ff8-4770-9846-870fbbcc35e0');

  final _uniqueId =
      M4eUniqueId.fromUniqueString('46e87882-4ff8-4770-9846-870fbbcc35e0');

  final _authToken = M4eAuthToken(id: _uniqueId, token: 'test token');

  final _tWallets = <M4eWallet>[
    M4eWallet(
      id: M4eUniqueId.fromUniqueString('46e87882-4ff8-4770-9846-870fbbcc35e0'),
      walletInfo: M4eWalletInfo(label: 'W1', currency: 'GHS'),
    ),
    M4eWallet(
      id: M4eUniqueId.fromUniqueString('9e8b5ecc-9f7e-4e06-ac3d-05b11a91da88'),
      walletInfo: M4eWalletInfo(label: 'W2', currency: 'GHS'),
    ),
  ];

  final _tWalletList = M4eWalletList(wallets: _tWallets);

  final _tM4eWalletForm =
      M4eWalletForm(label: 'test', currency: M4eCurrencyType.ghs());

  group('m4e-wallet-test | ', () {
    group('getWallets | ', () {
      test('should verify device connectivity status', () async {
        // arrange
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => true);
        when(_mockWalletApi.getUserWallets(any, any))
            .thenAnswer((_) async => _tWalletList);
        when(_mockAuthApi.getAuthTokenFromCache())
            .thenAnswer((_) async => _authToken);

        // act
        await _m4eWalletService.getWallets();

        // assert
        verify(_mockDataConnectionChecker.hasConnection);
      });

      group('isConnected', () {
        setUp(() {
          when(_mockDataConnectionChecker.hasConnection)
              .thenAnswer((_) async => true);
          when(_mockWalletApi.getUserWallets(any, any))
              .thenAnswer((_) async => _tWalletList);
          when(_mockAuthApi.getAuthTokenFromCache())
              .thenAnswer((_) async => _authToken);
        });

        test('should get [AuthToken] from cache', () async {
          // act
          await _m4eWalletService.getWallets();

          // assert
          verify(_mockAuthApi.getAuthTokenFromCache());
        });

        test('should return [M4eWalletList] on success call', () async {
          // arrange
          when(_mockWalletApi.getUserWallets(any, any))
              .thenAnswer((_) async => _tWalletList);

          // act
          final _result = await _m4eWalletService.getWallets();

          // assert
          expect(_result, equals(_tWalletList));
        });

        test('should re-throw a ServerException when server call fails',
            () async {
          // arrange
          when(_mockWalletApi.getUserWallets(any, any))
              .thenThrow(M4eExceptionMessages.kUndefinedException);

          // act
          final _call = _m4eWalletService.getWallets;

          // assert
          expect(() => _call(), throwsA(TypeMatcher<M4eServerException>()));
        });

        test('should throw an [AuthException] if no user is authenticated',
            () async {
          // arrange
          when(_mockAuthApi.getAuthTokenFromCache())
              .thenThrow(M4eExceptionMessages.kNoAuthUser);

          // act
          final _call = _m4eWalletService.getWallets;

          // assert
          expect(() => _call(), throwsA(TypeMatcher<M4eAuthException>()));
        });
      });

      group('isDisconnected', () {
        setUp(() {
          when(_mockDataConnectionChecker.hasConnection)
              .thenAnswer((_) async => false);
        });
        test(
            'should throw [ServerException] with [NO_INTERNET_CONNECTION] error code',
            () async {
          // arrange
          when(_mockWalletApi.getUserWallets(any, any))
              .thenThrow(M4eExceptionMessages.kNoInternetConnectionException);

          // act
          final _call = _m4eWalletService.getWallets;

          // assert
          expect(() => _call(), throwsA(TypeMatcher<M4eServerException>()));
        });
      });
    });

    group('createWallet | ', () {
      test('should verify device connectivity status', () async {
        // arrange
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => true);
        when(_mockAuthApi.getAuthTokenFromCache())
            .thenAnswer((_) async => _authToken);

        // act
        await _m4eWalletService.createWallet(_tM4eWalletForm);

        // assert
        verify(_mockDataConnectionChecker.hasConnection);
      });

      group('isConnected', () {
        setUp(() {
          when(_mockDataConnectionChecker.hasConnection)
              .thenAnswer((_) async => true);
          when(_mockAuthApi.getAuthTokenFromCache())
              .thenAnswer((_) async => _authToken);
        });

        test('should get [AuthToken] from cache', () async {
          // act
          await _m4eWalletService.getWallets();

          // assert
          verify(_mockAuthApi.getAuthTokenFromCache());
        });

        test('should call [M4eWalletApi] to createWallet with right parameters',
            () async {
          // assert
          when(_mockWalletApi.createWallet(any, any))
              .thenAnswer((_) async => _uniqueId);

          // act
          final _result = await _m4eWalletService.createWallet(_tM4eWalletForm);

          // assert
          verify(_mockWalletApi.createWallet(_authToken.id, _tM4eWalletForm));
          expect(_result, equals(_uniqueId));
        });

        test('should re-throw a ServerException when server call fails',
            () async {
          // arrange
          when(_mockWalletApi.createWallet(any, any))
              .thenThrow(M4eExceptionMessages.kUndefinedException);

          // act
          final _call = _m4eWalletService.createWallet;

          // assert
          expect(() => _call(_tM4eWalletForm),
              throwsA(TypeMatcher<M4eServerException>()));
        });

        test('should throw an [AuthException] if no user is authenticated',
            () async {
          // arrange
          when(_mockAuthApi.getAuthTokenFromCache())
              .thenThrow(M4eExceptionMessages.kNoAuthUser);

          // act
          final _call = _m4eWalletService.createWallet;

          // assert
          expect(() => _call(_tM4eWalletForm),
              throwsA(TypeMatcher<M4eAuthException>()));
        });
      });

      group('isDisconnected', () {
        setUp(() {
          when(_mockDataConnectionChecker.hasConnection)
              .thenAnswer((_) async => false);
        });
        test(
            'should throw [ServerException] with [NO_INTERNET_CONNECTION] error code',
            () async {
          // arrange
          when(_mockWalletApi.createWallet(any, any))
              .thenThrow(M4eExceptionMessages.kNoInternetConnectionException);

          // act
          final _call = _m4eWalletService.createWallet;

          // assert
          expect(() => _call(_tM4eWalletForm),
              throwsA(TypeMatcher<M4eServerException>()));
        });
      });
    });
  });
}
